#!/bin/bash

#===================
if [[ $EUID -ne 0 ]]; then
   echo "This script must be run as root like this:"
   echo "sudo $0"
   exit 1
fi
#===================
style="wide"
style="narrow"

encoding="ascii"
encoding="utf8"

#method="-H1" # -A intel-conf1
method=""    # default method which for macOS is -A darwin

#===================
if [ $encoding == "ascii" ]; then
	box='\-\+|.'
else
	box='┐─└├│┬'
fi

if [ $style == "wide" ]; then

	if [ $encoding == "ascii" ]; then
		# since we're using \, we want to indent one extra space
		rootindent='            '
		subindent='                '
	else
		rootindent='           '
		subindent='               '
	fi

	domainprefix="\\${box:1:1}\\["
	domainsuffix="\\]\\${box:1:1}\\${box:0:1}"

	bridgeprefix="\\${box:3:1}\\${box:1:1}"
	bridgesuffixlong="\\${box:1:1}\\${box:1:1}\\${box:1:1}\\${box:1:1}\\${box:0:1}"
	bridgesuffix="\\${box:1:1}\\${box:0:1}"

	deviceprefix="\\${box:3:1}\\${box:1:1}"
	devicesuffix=""
else
	rootindent=''
	subindent=' '

	domainprefix="\\${box:5:1}\\["
	domainsuffix="\\]"

	bridgeprefix="\\${box:3:1}\\${box:5:1}"
	bridgesuffixlong=""
	bridgesuffix=""

	deviceprefix="\\${box:3:1}\\${box:1:1}"
	devicesuffix=""
	
	fulladdress=1
fi

#===================

setpci $method -v -s '*:*:*' HEADER_TYPE.b PRIMARY_BUS.l 2> /dev/null | perl -0777 -pe '
	s/((.{4}):(..):(..\..)) \@0e = ([08][12])\n\g1 \@18 = ..(.{2})(.{2})(.{2})\n/$2:$3%$4 $2:$8 $2:$7 $2:$6\n/g; # bridge(01) or card bus(02) header type, multifunction(80) or not (00).
	s/((.{4}):(..):(..\..)) \@0e = (..)\n\g1 \@18 = ..(.{2})(.{2})(.{2})\n/$2:$3%$4\n/g # endpoint (or anything else) header type
' \
| sort -r > /tmp/pcilist.txt # get header type and bus information for all pci devices - reverse sort so last bus is first; format is "domain:bus%device.function domain:primary domain:secondary domain:subordinate"

# if primary = secondary = subordinate then that is a problem - change them to XX
sed '/^\(....:..%..\..\) \(\(.\{4\}:\)..\) \2 \2$/s//\1 \3XX \3XX \3XX/' /tmp/pcilist.txt > /tmp/pcilist2.txt

echo '/ ....:/s// /g' > /tmp/pcipaths.txt # add a command that will remove domain from primary, secondary, and subordinate bus numbers
sed -nE '/^(....:..%..\..) .{4}:.. (.{4}:..) .*/s//\/(\2)\/s\/\/\1%\\1\/g/p' /tmp/pcilist2.txt >> /tmp/pcipaths.txt # create device path substitution commands that will add parent bus to each secondary bus
sed -E -f /tmp/pcipaths.txt /tmp/pcilist2.txt > /tmp/pcitree1.txt # apply the commands
sed -E '/^(....:..).*/s//\1/g' /tmp/pcitree1.txt | sort -u > /tmp/pcitree2.txt # get all root complexes (usually just 0000:00)

cat /tmp/pcitree1.txt /tmp/pcitree2.txt | sort > /tmp/pcitree3.txt # concatenate the devices and root complexes

# use perl to do some formatting
perl -CSDA -Mutf8 -pe '
	s/^(.*(\w\w\w\w:\w\w%\w\w\.\w).*)/\1  # \2/;                                             # save domain/bus/device/function after a "  # " - this will be used to append name and ids
	s/^\w\w\w\w:\w\w%/'"$rootindent"'/;                                                      # remove domain/bus number from left side and indent
	s/\w\w\.\w%\w\w\w\w:\w\w%/'"$subindent"'/g;                                              # use the paths to do more indenting
	s/(\w\w\.\w) \w\w (\w\w) \g2/'"${bridgeprefix}"'\1-[\2]'"${bridgesuffixlong}"'/;         # format the secondary and subordinate bus numbers for single bus bridge
	s/(\w\w\.\w) \w\w (\w\w) (\w\w)/'"${bridgeprefix}"'\1-[\2-\3]'"${bridgesuffix}"'/;       # format the secondary and subordinate bus numbers for multiple bus bridge
	s/%/:/;                                                                                  # replace % bus/device delimiter with standard ':'
	s/(\w\w\.\w  )/'"${deviceprefix}"'\1'"${devicesuffix}"'/;                                # add the "+-" formating for devices
	s/^(\w\w\w\w:\w\w)$/'"${domainprefix}"'\1'"${domainsuffix}"'/                            # add the -[ ]\" for domain/bus
' \
/tmp/pcitree3.txt > /tmp/pcitree4.txt

perl -CSDA -Mutf8 -0777 -pe '
	#	add | between devices of the same bus
	do {
		$good=0;
		while (
			/											#	begin regular expression
				(										# 		begin g1
					\n									#			\n
					([\ '"\\${box:4:1}"']*)				#			[ |]* -> g2
					['"\\${box:3:1}\\${box:4:1}"']		#			[+|]
					[^\n]*+								#			rest of the line up to \n, give nothing back
					(?{$X=pos()})						#			mark position
					\n									#			\n
				)										#		end g1
				(?=										#		begin lookahead
					(									#			begin group
						\g2\ [^\n]*+					#				following line(s) start with g2 also
						\n								#
					)++									#			end group, match one or more, give nothing back
					\g2									#			the line after those start with g2
					'"\\${box:3:1}"'					#				 and "+" meaning that another device follows g1 on the same bus
				)										#		end lookahead
				(
					\g2									#
					(?{$C=pos()})						#		$C is the position where we want to insert the |
					\ 									#		the character at position $C to be modified is a space character
				)										#
			/gx											#	end regular expression
		) {
			$good = 1;
			substr($_,$C,1) = "'"\\${box:4:1}"'";		#	replace the space character at position $C with a |
			pos() = $X;									#	start next search at $X which is the end of the previous line
		}
	} until ($good == 0);

	# replace + with \ for last device of a bus
	s/(\n[\ '"\\${box:4:1}"']*)'"\\${box:3:1}"'(?![^\n]*\g1['"\\${box:3:1}\\${box:4:1}"'])/\1'"\\${box:2:1}"'/g;
' /tmp/pcitree4.txt > /tmp/pcitree5.txt

# get the length of the longest line so we can align device information in a detail column
spaces=$(tr -c "\n" " " < /tmp/pcitree5.txt | sort -r | head -n 1)

lspci $method -nn -D 2> /dev/null | sed -E '/\//s//\\\//g;/^(....:..:..\..) ([^[]*) (\[....\])(.*) (\[....:....\])(( \(rev ..\)))?/s//\/# (\1)$\/s\/\/'"$spaces"'#                \\1 \5 \3\7 \2\4\//' > /tmp/pcidevices.txt
echo '/^(.{'"$((${#spaces} - 14))"'}) *(.*)/s//\1\2/' >> /tmp/pcidevices.txt # create the substitutions for the detail column

# get link width and speed max/current
pat="(....:..:..\..) @00 = (....)(....)"
IFS=$'\n'
echo "" > /tmp/pcidevices2.txt
for thedevice in $(sudo setpci $method -v -s '*:*:*' VENDOR_ID.l 2> /dev/null); do
	if [[ $thedevice =~ $pat ]]; then
		pcidevice=${BASH_REMATCH[1]}
		LinkRegisters=$(sudo setpci $method -s $pcidevice CAP_EXP+12.w CAP_EXP+c.l 2> /dev/null)
		if [[ -n $LinkRegisters ]]; then
			LinkStatusRegister=${LinkRegisters:0:4}
			LinkWidth=$((0x$LinkStatusRegister >> 4 & 31))
			if [[ $LinkWidth -ge 0 ]]; then
				LinkSpeed=$((0x$LinkStatusRegister & 15))
				LinkCapabilitiesRegister=${LinkRegisters:5:8}
				MaxLinkWidth=$((0x$LinkCapabilitiesRegister >> 4 & 31))
				MaxLinkSpeed=$((0x$LinkCapabilitiesRegister & 15))
				echo "/#                ($pcidevice .*$)/s//# g${MaxLinkSpeed}x${MaxLinkWidth} > g${LinkSpeed}x${LinkWidth}                 \1/" >> /tmp/pcidevices2.txt
			fi
		fi
	else
		echo "# Unexpected pattern $thedevice"
	fi
done

perl -CSDA -Mutf8 -pe '
s/# (g.x..?) > \g1 /# \1     /;
s/(\/s\/\/#.{16}) */\1/;
'  -i /tmp/pcidevices2.txt

# if there's only one domain/segment then remove them from the output
if [[ $(wc -l < /tmp/pcitree2.txt) -eq 1 ]]; then
	echo "/(#.{16}).{4}:/s//\\1/" >> /tmp/pcidevices2.txt
fi

if [[ $fulladdress -eq 1 ]]; then
	echo "/..\..([^#]*#.{16})([0-9a-f.:]+) /s//\\2\\1/" >> /tmp/pcidevices2.txt
fi


sed -E -f /tmp/pcidevices.txt -f /tmp/pcidevices2.txt /tmp/pcitree5.txt > /tmp/pcitree6.txt # apply the substitutions

echo '#========================================================================================='
cat /tmp/pcitree6.txt # final output
